/* drive.c
 * Designed to run on Create Command Module by Matt Mohorn and Daniel Bernstein
 *
 * The basic architecture of this program can be re-used to easily 
 * write a wide variety of Create control programs.  All sensor values
 * are polled in the background (using the serial rx interrupt) and 
 * stored in the sensors array as long as the function 
 * delayAndUpdateSensors() is called periodically.  Users can send commands
 * directly a byte at a time using byteTx() or they can use the 
 * provided functions, such as baud() and drive().
 */



// Includes
#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdlib.h>
#include "oi.h"


// Constants
#define RESET_SONG 0
#define START_SONG 1
#define BUMP_SONG  2
#define END_SONG   3

#define HALF_DOOR_LENGTH 700
#define QUARTER_DOOR_LENGTH 330
#define NUM_DOORS 8
#define DEFAULT_SPEED 100

#define NUM_DOORS_EAST 2

#define USB 1
#define CR8 2



// Global variables
volatile uint16_t timer_cnt = 0;
volatile uint8_t timer_on = 0;
volatile uint8_t sensors_flag = 0;
volatile uint8_t sensors_index = 0;
volatile uint8_t sensors_in[Sen6Size];
volatile uint8_t sensors[Sen6Size];
volatile int16_t distance = 0;
volatile int16_t angle = 0;
volatile int16_t roomDepth = HALF_DOOR_LENGTH+HALF_DOOR_LENGTH; //how far into the room to travel
volatile int16_t waitTime = 48000; // time in ms to wait in the room
volatile int16_t wiggleRad = 100;





// Functions
void byteTx(uint8_t value);
void delayMs(uint16_t time_ms);
void delayAndUpdateSensors(unsigned int time_ms);
void initialize(void);
void powerOnRobot(void);
void baud(uint8_t baud_code);
void drive(int16_t velocity, int16_t radius);
uint16_t divide(uint16_t x, uint16_t y);
void defineSongs(void);
void playResetSong(void);
void playStartSong(void);
void goFeet(uint16_t howMany);
void turnNinetyDegrees(int8_t direction);
void turnInPlace(int8_t direction);
void goDistance(int16_t distance);
void followWall(void);
void faceStraight(void);
void wiggleWall(void);
void visitRoom(void);


int main (void) 
{
    // Set up Create and module
    initialize();
    LEDBothOff;
    powerOnRobot();
    byteTx(CmdStart);
    baud(Baud19200);
    defineSongs();
    byteTx(CmdControl);
    byteTx(CmdFull);

    // Stop just as a precaution
    drive(0, RadStraight);
    delayAndUpdateSensors(20);
    playResetSong();
    
    
    //Set the distance to each door  Room #   {3068, 3084, 3106  3196, 3198, 3234, 3209, 3187}
    int16_t distanceToDoor[NUM_DOORS] =       {6400, 600,  4200, 3660, 9600, 11000, 700, 8300};

    
 /* 
    
    for(;;){
        
        //DDRB = 0xF8; // set pin B4 (already an output)
        DDRC = 0xF;//set direction of middle pins to all outputs
        PORTC |= 0x04; 
        
        DDRB = 0xF;//set direction of middle pins to all outputs
        PORTB |= 0x04; 
        
        DDRD = 0xF;//set direction of middle pins to all outputs
        PORTD |= 0x04; 
        
        
        if(UserButtonPressed){
            playResetSong();
            
            
            
            
            
            byteTx(147);
            byteTx(255);
        } // end if
        delayAndUpdateSensors(20);
    } // end for
 */ 
 
 
    
    for(;;){
        delayAndUpdateSensors(20);
        
        //if (sensors[SenVWall]) playResetSong();
        
        int16_t distanceTraveled = 0;
        int16_t distanceTraveled2 = 0;
        int16_t angleFacing = 0;
        int8_t currentDoor = 0;
        
        if(UserButtonPressed){
            
            playStartSong();
            
            //navigate until the black button is pressed or all doors have been visited
            while(!UserButtonPressed && currentDoor < NUM_DOORS){
                
                /*
                
                while(!sensors[SenVWall]){ //continue in middle of hallway until a vwall is sensed
                    drive(200,RadStraight);
                    delayAndUpdateSensors(20);
                    
                }//Virtual wall as been sensed
                
                while(!(sensors[SenBumpDrop] & BumpEither)){//wiggle till roomba hits the wall
                    if (sensors[SenVWall]) 
                        drive(100, wiggleRad); // arc left away from virtual wall
                    else 
                        drive(100, -wiggleRad); // arc right toward the wall
                    delayAndUpdateSensors(20);
                } // end while !bump
                */
                
                
                followWall();
                distanceTraveled += (((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
                
                
                //we are at a door
                if(distanceTraveled >= distanceToDoor[currentDoor] && !sensors[SenWall]){
                    distanceTraveled = 0;
                    drive(0,RadStraight);
//                    playResetSong();
                    goDistance(HALF_DOOR_LENGTH);
                    turnNinetyDegrees(right);
		    
                   visitRoom();
            
                    //turnNinetyDegrees(left);
                    
                    if(currentDoor==5){ //after last door on right we will bail and head back
                        goDistance(1500);
                        drive(0,RadStraight);
                        turnNinetyDegrees(left);
                        
                        while(!sensors[SenVWall]){ //continue in middle of hallway until a vwall is sensed
                            drive(200,RadStraight);
                            delayAndUpdateSensors(20);
                            
                        }//Virtual wall as been sensed
                        
                        
                        while(!(sensors[SenBumpDrop] & BumpEither)){//wiggle till roomba hits the wall
                            if (sensors[SenVWall]) 
                                drive(100, wiggleRad); // arc left away from virtual wall
                            else 
                                drive(100, -wiggleRad); // arc right toward the wall
                            delayAndUpdateSensors(20);
                        } // end while !bump
                         
                    }
                    else if(currentDoor==1){
                        int16_t dis = 0;
                        while(!(sensors[SenBumpDrop] & BumpEither)){//arc away from door until bumped
                            drive(DEFAULT_SPEED, -QUARTER_DOOR_LENGTH);
                            delayAndUpdateSensors(20);
                        } // end while !bump
                        faceStraight();
                        while (dis<5000) {
                            followWall();
                            delayAndUpdateSensors(20);
                            dis+=(((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
                        }
                        drive(0,RadStraight); //stop
                        turnNinetyDegrees(left);
                        goDistance(1000);
                        drive(0,RadStraight);
                        turnNinetyDegrees(right);
                        
                        drive(200,RadStraight);
                        
                        while(!sensors[SenVWall]){ //continue in middle of hallway until a vwall is sensed
                            drive(200,RadStraight);
                            delayAndUpdateSensors(20);
                            
                        }//Virtual wall as been sensed
                        
                        while(!(sensors[SenBumpDrop] & BumpEither)){//wiggle till roomba hits the wall
                            if (sensors[SenVWall]) 
                                drive(100, wiggleRad); // arc left away from virtual wall
                            else 
                                drive(100, -wiggleRad); // arc right toward the wall
                            delayAndUpdateSensors(20);
                        } // end while !bump
                        
                    }
                    else if(currentDoor==2){
                        int16_t dis = 0;
                        while(!(sensors[SenBumpDrop] & BumpEither)){//arc away from door until bumped
                            drive(DEFAULT_SPEED, -HALF_DOOR_LENGTH);
                            delayAndUpdateSensors(20);
                        } // end while !bump
                        faceStraight();
                        while (dis<1500) {
                            followWall();
                            delayAndUpdateSensors(20);
                            dis+=(((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
                        }
                        drive(0,RadStraight); //stop
                        turnNinetyDegrees(left);
                        goDistance(1000);
                        drive(0,RadStraight);
                        turnNinetyDegrees(right);
                        
                        while(!sensors[SenVWall]){ //continue in middle of hallway until a vwall is sensed
                            drive(200,RadStraight);
                            delayAndUpdateSensors(20);
                            
                        }//Virtual wall as been sensed
                        
                        while(!(sensors[SenBumpDrop] & BumpEither)){//wiggle till roomba hits the wall
                            if (sensors[SenVWall]) 
                                drive(100, 200); // arc left away from virtual wall
                            else 
                                drive(100, -200); // arc right toward the wall
                            delayAndUpdateSensors(20);
                        } // end while !bump
                    }
                    else{
                        while(!(sensors[SenBumpDrop] & BumpEither)){//arc away from door until bumped
                            drive(DEFAULT_SPEED, -QUARTER_DOOR_LENGTH);
                            delayAndUpdateSensors(20);
                        } // end while !bump
                        
                    }

                    
                    while((sensors[SenBumpDrop] & BumpEither) || !sensors[SenWall]){
                        turnInPlace(left);
                        delayAndUpdateSensors(20);
                    } // end while bump
		    
                    faceStraight();
                    //playResetSong();
                    //angleFacing = 0;
                    currentDoor++;
                } // end if
            } // end while(!UserButtonPressed)
            
            
        
            drive(0,RadStraight);
            delayAndUpdateSensors(120);
            
           
            
            
            
        } // end if(UserButtonPressed)
    } // end for(;;)
     
     
} // end main


void visitRoom(){ 
    int16_t depth= roomDepth ;//depth is how far into the room it goes

    while (depth>=0 && ! ((sensors[SenBumpDrop] & BumpEither))) {
        if(depth>(roomDepth - 500)) //EDIT when depth is changed
            drive(300,RadStraight);//give the robot a little kick over the threshold
        
        else
            drive(100,RadStraight);
        
        delayAndUpdateSensors(20);
        depth-=(((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
    }
    
    drive(0,RadStraight); //stop

    
    if(depth<=0){//we bumped the door because it was closed
        /*
         *The roomba has entered the room. Take some pictures.
         */
        
        delayAndUpdateSensors(waitTime);  
    }
    turnNinetyDegrees(left);        
    turnNinetyDegrees(left);// run away
    
    while(depth<roomDepth){
        drive(100,RadStraight);
        delayAndUpdateSensors(20);
        depth+=(((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
        
    }
    
    
    
}

void wiggleWall(){
    if (sensors[SenVWall]) 
        drive(100, 200); // arc left away from virtual wall
    else 
        drive(100, -200); // arc right toward the wall
    
}//end wiggle


void faceStraight(){
    turnInPlace(left);
    delayAndUpdateSensors(380);
    drive(0,RadStraight);
} // end faceStraight



void goDistance(int16_t distance){
    while(distance > 0){
        drive(200, RadStraight);
        delayAndUpdateSensors(20);
        distance -= (((int16_t)sensors[SenDist1]) << 8) + ((int16_t) sensors[SenDist0]);
    } // end while(!UserButtonPressed)
} // end goDistance



void followWall(void){
    if(sensors[SenBumpDrop] & BumpEither)
        turnInPlace(left);
    else if(!sensors[SenWall])
        drive(DEFAULT_SPEED,-200);
    else
        drive(DEFAULT_SPEED, RadStraight);
        delayAndUpdateSensors(20);
} // end followWall

void turnInPlace(int8_t direction){
    int16_t velocity = DEFAULT_SPEED;
    int16_t rightWheelVelocity = -1 * direction * velocity;
    int16_t leftWheelVelocity = direction * velocity;
    byteTx(CmdDriveWheels);
    byteTx((rightWheelVelocity >> 8) & 255);
    byteTx(rightWheelVelocity & 255);
    byteTx((leftWheelVelocity >> 8) & 255);
    byteTx(leftWheelVelocity & 255);
} // end turnInPlace

void goFeet(uint16_t howMany){ // go howMany feet
    uint16_t i;
    drive(320, RadStraight);
    for(i = 0; i<howMany; i++){
	delayAndUpdateSensors(1000);
    } // end for
    drive(0,RadStraight);
} // end goDistance

void turnNinetyDegrees(int8_t direction){
    turnInPlace(direction);
    delayAndUpdateSensors(2090);
    drive(0,RadStraight);
} // end turnNinetyDegrees

// does division using shifts
uint16_t divide(uint16_t x, uint16_t y){
    uint16_t a, b, q, counter;
    q = 0;
    while(x >= y){
        a = x >> 1;
        b = y;
        counter = 1;
        while(a >= b){
	    b <<= 1;
            counter <<= 1;
        } // end while
        x -= b;
        q += counter;
    } // end while
    return q;
} // end divide


// Serial receive interrupt to store sensor values
SIGNAL(SIG_USART_RECV)
{
  uint8_t temp;


  temp = UDR0;

  if(sensors_flag)
  {
    sensors_in[sensors_index++] = temp;
    if(sensors_index >= Sen6Size)
      sensors_flag = 0;
  }
}

void playResetSong(void){
  byteTx(CmdPlay);
  byteTx(RESET_SONG);
  delayAndUpdateSensors(750);
} // end playResetSongAndWait



// Timer 1 interrupt to time delays in ms
SIGNAL(SIG_OUTPUT_COMPARE1A)
{
  if(timer_cnt)
    timer_cnt--;
  else
    timer_on = 0;
}




// Transmit a byte over the serial port
void byteTx(uint8_t value)
{
  while(!(UCSR0A & _BV(UDRE0))) ;
  UDR0 = value;
}




// Delay for the specified time in ms without updating sensor values
void delayMs(uint16_t time_ms)
{
  timer_on = 1;
  timer_cnt = time_ms;
  while(timer_on) ;
}




// Delay for the specified time in ms and update sensor values
void delayAndUpdateSensors(uint16_t time_ms)
{
  uint8_t temp;

  timer_on = 1;
  timer_cnt = time_ms;
  while(timer_on)
  {
    if(!sensors_flag)
    {
      for(temp = 0; temp < Sen6Size; temp++)
        sensors[temp] = sensors_in[temp];

      // Update running totals of distance and angle
      distance += (int)((sensors[SenDist1] << 8) | sensors[SenDist0]);
      angle += (int)((sensors[SenAng1] << 8) | sensors[SenAng0]);

      byteTx(CmdSensors);
      byteTx(6);
      sensors_index = 0;
      sensors_flag = 1;
    }
  }
}




// Initialize the Mind Control's ATmega168 microcontroller
void initialize(void)
{
  cli();

  // Set I/O pins
  DDRB = 0x10;
  PORTB = 0xCF;
  
    
  //all of the middle serial port pins are inputs  
  DDRC = 0x00;
  PORTC = 0xFF;
    
  DDRD = 0xE6;
  PORTD = 0x7D;

  // Set up timer 1 to generate an interrupt every 1 ms
  TCCR1A = 0x00;
  TCCR1B = (_BV(WGM12) | _BV(CS12));
  OCR1A = 71;
  TIMSK1 = _BV(OCIE1A);

  // Set up the serial port with rx interrupt
  UBRR0 = 19;
  UCSR0B = (_BV(RXCIE0) | _BV(TXEN0) | _BV(RXEN0));
  UCSR0C = (_BV(UCSZ00) | _BV(UCSZ01));

  // Turn on interrupts
  sei();
}




void powerOnRobot(void)
{
  // If Create's power is off, turn it on
  if(!RobotIsOn)
  {
      while(!RobotIsOn)
      {
          RobotPwrToggleLow;
          delayMs(500);  // Delay in this state
          RobotPwrToggleHigh;  // Low to high transition to toggle power
          delayMs(100);  // Delay in this state
          RobotPwrToggleLow;
      }
      delayMs(3500);  // Delay for startup
  }
}




// Switch the baud rate on both Create and module
void baud(uint8_t baud_code)
{
  if(baud_code <= 11)
  {
    byteTx(CmdBaud);
    UCSR0A |= _BV(TXC0);
    byteTx(baud_code);
    // Wait until transmit is complete
    while(!(UCSR0A & _BV(TXC0))) ;

    cli();

    // Switch the baud rate register
    if(baud_code == Baud115200)
      UBRR0 = Ubrr115200;
    else if(baud_code == Baud57600)
      UBRR0 = Ubrr57600;
    else if(baud_code == Baud38400)
      UBRR0 = Ubrr38400;
    else if(baud_code == Baud28800)
      UBRR0 = Ubrr28800;
    else if(baud_code == Baud19200)
      UBRR0 = Ubrr19200;
    else if(baud_code == Baud14400)
      UBRR0 = Ubrr14400;
    else if(baud_code == Baud9600)
      UBRR0 = Ubrr9600;
    else if(baud_code == Baud4800)
      UBRR0 = Ubrr4800;
    else if(baud_code == Baud2400)
      UBRR0 = Ubrr2400;
    else if(baud_code == Baud1200)
      UBRR0 = Ubrr1200;
    else if(baud_code == Baud600)
      UBRR0 = Ubrr600;
    else if(baud_code == Baud300)
      UBRR0 = Ubrr300;

    sei();

    delayMs(100);
  }
}



// Send Create drive commands in terms of velocity and radius
void drive(int16_t velocity, int16_t radius)
{
  byteTx(CmdDrive);
  byteTx((uint8_t)((velocity >> 8) & 0x00FF));
  byteTx((uint8_t)(velocity & 0x00FF));
  byteTx((uint8_t)((radius >> 8) & 0x00FF));
  byteTx((uint8_t)(radius & 0x00FF));
}



// Define songs to be played later
void defineSongs(void)
{
  // Reset song
  byteTx(CmdSong);
  byteTx(RESET_SONG);
  byteTx(4);
  byteTx(60);
  byteTx(6);
  byteTx(72);
  byteTx(6);
  byteTx(84);
  byteTx(6);
  byteTx(96);
  byteTx(6);

  // Start song
  byteTx(CmdSong);
  byteTx(START_SONG);
  byteTx(6);
  byteTx(69);
  byteTx(18);
  byteTx(72);
  byteTx(12);
  byteTx(74);
  byteTx(12);
  byteTx(72);
  byteTx(12);
  byteTx(69);
  byteTx(12);
  byteTx(77);
  byteTx(24);

  // Bump song
  byteTx(CmdSong);
  byteTx(BUMP_SONG);
  byteTx(2);
  byteTx(74);
  byteTx(12);
  byteTx(59);
  byteTx(24);

  // End song
  byteTx(CmdSong);
  byteTx(END_SONG);
  byteTx(5);
  byteTx(77);
  byteTx(18);
  byteTx(74);
  byteTx(12);
  byteTx(72);
  byteTx(12);
  byteTx(69);
  byteTx(12);
  byteTx(65);
  byteTx(24);
}

void playStartSong(void){
    byteTx(CmdPlay);
    byteTx(START_SONG);
    delayAndUpdateSensors(2813);
} // end playStartSong

void setSerial(uint8_t com) {
    if(com == USB)
        PORTB |= 0x10;
    else if(com == CR8)
        PORTB &= ~0x10;
}

uint8_t getSerialDestination(void) {
    if (PORTB & 0x10)
        return USB;
    else
        return CR8;
}
/*

void writeChar(char c, uint8_t com) {
    uint8_t originalDestination = getSerialDestination();
    if (com != originalDestination) {
        setSerial(com);
        delayus(200);
    }
    byteTx((uint8_t)(c));
    if (com != originalDestination) {
        setSerial(originalDestination);
        delayus(200);
    }
}
*/
